from typing import Optional
from pathlib import Path

import smbus2
from typed_registers import SMBusRegisterBus

from bussdcc import Device

from .config import I2CBusConfig


class I2CBus(Device[I2CBusConfig]):
    kind = "bus"

    def __init__(self, *, id: str, config: I2CBusConfig):
        super().__init__(id=id, config=config)
        self._bus: smbus2.SMBus | None = None

    def connect(self) -> None:
        self._bus = smbus2.SMBus(self.config.bus)

    def disconnect(self) -> None:
        if self._bus:
            self._bus.close()
            self._bus = None

    def protocol(self) -> Optional[SMBusRegisterBus]:
        if not self._bus:
            return None

        return SMBusRegisterBus(self._bus)

    def discover_devices(self) -> list[int]:
        if not self._bus:
            return []

        found = []
        for addr in range(0x03, 0x78):
            try:
                # Read a single byte without specifying a register
                msg = smbus2.i2c_msg.read(addr, 1)
                self._bus.i2c_rdwr(msg)
                found.append(addr)
            except OSError:
                pass
        return found

    @classmethod
    def discover_buses(cls) -> list[int]:
        buses = []

        for path in Path("/dev").glob("i2c-*"):
            try:
                buses.append(int(path.name.split("-")[1]))
            except (IndexError, ValueError):
                continue

        return sorted(buses)



from typing import Protocol, Literal, runtime_checkable

PinDirection = Literal["in", "out"]
Pull = Literal["up", "down", "off"]
Level = Literal[0, 1]


@runtime_checkable
class GPIOInterface(Protocol):
    def setup(
        self,
        pin: int,
        direction: PinDirection,
        *,
        pull: Pull = "off",
        initial: Level | None = None,
    ) -> None: ...
    def input(self, pin: int) -> Level: ...
    def output(self, pin: int, value: Level) -> None: ...



from bussdcc.device.definition import DeviceDefinition

from .config import USBCameraConfig
from .driver import USBCamera

definition = DeviceDefinition(
    config_class=USBCameraConfig,
    driver_class=USBCamera,
)



from typing import Any, Literal
from dataclasses import dataclass, field


@dataclass(slots=True)
class NAU7802CalibrationConfig:
    offset: int = field(
        default=0,
        metadata={
            "label": "Offset",
            "group": "Calibration",
            "ui": "number",
        },
    )

    scale: float = field(
        default=1.0,
        metadata={
            "label": "Scale Factor",
            "group": "Calibration",
            "ui": "number",
            "step": 0.0001,
        },
    )

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "NAU7802CalibrationConfig":
        return cls(
            offset=data.get("offset", 0),
            scale=data.get("scale", 1.0),
        )


@dataclass(slots=True)
class NAU7802ChannelConfig:
    calibration: NAU7802CalibrationConfig = field(metadata={"group": "Channels"})

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "NAU7802ChannelConfig":
        return cls(
            calibration=NAU7802CalibrationConfig.from_dict(data.get("calibration", {}))
        )


@dataclass(slots=True)
class NAU7802Config:
    bus_id: str = field(
        metadata={
            "label": "I²C Bus",
            "group": "Connection",
            "ui": "bus",
        }
    )

    addr: int = field(
        default=0x2A,
        metadata={
            "label": "I²C Address",
            "group": "Connection",
            "ui": "hex",
            "min": 0,
            "max": 127,
        },
    )

    gain: Literal[1, 2, 4, 8, 16, 32, 64, 128] = field(
        default=128,
        metadata={
            "label": "Gain",
            "group": "ADC",
            "ui": "select",
            "options": [1, 2, 4, 8, 16, 32, 64, 128],
        },
    )

    sample_rate: Literal[10, 20, 40, 80, 320] = field(
        default=10,
        metadata={
            "label": "Sample Rate (SPS)",
            "group": "ADC",
            "ui": "select",
            "options": [10, 20, 40, 80, 320],
        },
    )

    samples: int = field(
        default=1,
        metadata={
            "label": "Samples per Reading",
            "group": "Performance",
            "ui": "number",
            "min": 1,
            "max": 64,
        },
    )

    discard_samples: int = field(
        default=8,
        metadata={
            "label": "Discard Samples After Channel Switch",
            "group": "Performance",
            "ui": "number",
            "min": 0,
            "max": 32,
        },
    )

    channels: dict[int, NAU7802ChannelConfig] = field(
        default_factory=dict,
        metadata={"group": "Channels"},
    )

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "NAU7802Config":
        channels = {
            int(k): NAU7802ChannelConfig.from_dict(v)
            for k, v in data.get("channels", {}).items()
        }

        return cls(
            bus_id=data["bus_id"],
            addr=data.get("addr", 0x2A),
            gain=data.get("gain", 128),
            sample_rate=data.get("sample_rate", 10),
            samples=data.get("samples", 1),
            discard_samples=data.get("discard_samples", 8),
            channels=channels,
        )



from typing import Any, Literal
from dataclasses import dataclass, field

SafeState = Literal["off", "on"]
Logic = Literal["active_high", "active_low"]


@dataclass(slots=True)
class DigitalOutputConfig:
    bus_id: str = field(
        metadata={
            "label": "GPIO Bus",
            "group": "Connection",
            "ui": "bus",
        }
    )

    pin: int = field(
        metadata={
            "label": "GPIO Output Pin",
            "group": "Hardware",
            "ui": "number",
            "min": 2,
            "max": 27,
        },
    )

    logic: Logic = field(
        default="active_high",
        metadata={
            "label": "Logic Level",
            "group": "Behavior",
            "ui": "select",
            "options": ["active_high", "active_low"],
            "help": "Determines whether the device is active when the pin is HIGH or LOW",
        },
    )

    safe_state: SafeState = field(
        default="off",
        metadata={
            "label": "Safe State",
            "group": "Safety",
            "ui": "select",
            "options": ["off", "on"],
            "help": "State to enter when device disconnects",
        },
    )

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "DigitalOutputConfig":
        return cls(
            bus_id=data["bus_id"],
            pin=data["pin"],
            logic=data.get("logic", "active_high"),
            safe_state=data.get("safe_state", "off"),
        )



from typing import Any, Literal
from dataclasses import dataclass, field

PinMode = Literal["BCM", "BOARD"]


@dataclass(slots=True)
class GPIOBusConfig:
    mode: PinMode = field(
        default="BCM",
        metadata={
            "label": "Pin Numbering Mode",
            "group": "GPIO",
            "ui": "select",
            "options": ["BCM", "BOARD"],
        },
    )

    warnings: bool = field(
        default=False,
        metadata={
            "label": "Enable GPIO Warnings",
            "group": "GPIO",
        },
    )

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "GPIOBusConfig":
        return cls(
            mode=data.get("mode", "BCM"),
            warnings=data.get("warnings", False),
        )



from bussdcc.device.definition import DeviceDefinition

from .config import GPIOBusConfig
from .driver import GPIOBus

definition = DeviceDefinition(
    config_class=GPIOBusConfig,
    driver_class=GPIOBus,
)



from typing import Literal

import RPi.GPIO as GPIO

from bussdcc import Device

from .config import GPIOBusConfig, PinMode
from .interface import GPIOInterface, PinDirection, Pull, Level

GPIOMode = Literal[10, 11]
GPIODirection = Literal[0, 1]


class GPIOBus(Device[GPIOBusConfig], GPIOInterface):
    kind = "bus"

    MODE_MAP: dict[PinMode, GPIOMode] = {
        "BCM": GPIO.BCM,
        "BOARD": GPIO.BOARD,
    }

    DIR_MAP: dict[PinDirection, GPIODirection] = {
        "in": GPIO.IN,
        "out": GPIO.OUT,
    }

    PULL_MAP: dict[Pull, int] = {
        "off": GPIO.PUD_OFF,
        "up": GPIO.PUD_UP,
        "down": GPIO.PUD_DOWN,
    }

    def connect(self) -> None:
        GPIO.setmode(self.MODE_MAP[self.config.mode])
        GPIO.setwarnings(self.config.warnings)

    def disconnect(self) -> None:
        GPIO.cleanup()

    def protocol(self) -> GPIOInterface:
        return self

    def setup(
        self,
        pin: int,
        direction: PinDirection,
        *,
        pull: Pull = "off",
        initial: Level | None = None,
    ) -> None:

        kwargs = {
            "pull_up_down": self.PULL_MAP[pull],
        }

        if initial is not None:
            kwargs["initial"] = GPIO.HIGH if initial else GPIO.LOW

        GPIO.setup(pin, self.DIR_MAP[direction], **kwargs)

    def input(self, pin: int) -> Level:
        return 1 if GPIO.input(pin) else 0

    def output(self, pin: int, value: Level) -> None:
        GPIO.output(pin, GPIO.HIGH if value else GPIO.LOW)



from bussdcc.device.definition import DeviceDefinition

from .config import W1BusConfig
from .driver import W1Bus

definition = DeviceDefinition(
    config_class=W1BusConfig,
    driver_class=W1Bus,
)



from typing import Any
from dataclasses import dataclass, field


@dataclass(slots=True)
class I2CBusConfig:
    bus: int = field(
        default=1,
        metadata={
            "label": "I²C Bus Number",
            "group": "Connection",
            "ui": "number",
            "min": 0,
        },
    )

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "I2CBusConfig":
        return cls(bus=data.get("bus", 1))



from bussdcc.device.definition import DeviceDefinition

from .config import DigitalOutputConfig
from .driver import DigitalOutput

definition = DeviceDefinition(
    config_class=DigitalOutputConfig,
    driver_class=DigitalOutput,
)



from bussdcc.device.definition import DeviceDefinition

from .config import I2CBusConfig
from .driver import I2CBus

definition = DeviceDefinition(
    config_class=I2CBusConfig,
    driver_class=I2CBus,
)



from typing import Any
from dataclasses import dataclass, field


@dataclass(slots=True)
class DS18B20Config:
    bus_id: str = field(
        metadata={
            "label": "1-Wire Bus",
            "group": "Connection",
            "ui": "bus",
        }
    )

    device_id: str = field(
        metadata={
            "label": "Device ID",
            "group": "Connection",
            "ui": "text",
        }
    )

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "DS18B20Config":
        return cls(bus_id=data["bus_id"], device_id=data["device_id"])



from typing import Any, Optional
import threading

import cv2

from bussdcc import Device

from .config import USBCameraConfig


class USBCamera(Device[USBCameraConfig]):
    kind = "camera"

    MAX_CONSECUTIVE_FAILURES = 3

    def __init__(self, *, id: str, config: USBCameraConfig):
        super().__init__(id=id, config=config)
        self._lock = threading.Lock()
        self._consecutive_failures = 0
        self.cap: cv2.VideoCapture | None = None

    def _str_to_fourcc(self, val: str) -> int:
        if val == "AUTO":
            return 0

        if len(val) != 4:
            raise ValueError("FOURCC must be exactly 4 characters")

        return (
            ord(val[0]) | (ord(val[1]) << 8) | (ord(val[2]) << 16) | (ord(val[3]) << 24)
        )

    def _fourcc_to_str(self, val: float) -> str:
        if val == 0:
            return "AUTO"

        h = int(val)
        return (
            chr(h & 0xFF)
            + chr((h >> 8) & 0xFF)
            + chr((h >> 16) & 0xFF)
            + chr((h >> 24) & 0xFF)
        )

    def _get_exposure(self) -> Any:
        if self.cap is None:
            return

        backend = self.get_backend_name()
        if backend == "V4L2":  # Linux
            return self.cap.get(cv2.CAP_PROP_AUTO_EXPOSURE) == 3.0
        elif backend == "AVFOUNDATION":  # macOS
            return self.cap.get(cv2.CAP_PROP_AUTO_EXPOSURE) == 0.0
        else:
            return self.cap.get(cv2.CAP_PROP_AUTO_EXPOSURE) == 3.0

    def _apply_config(self) -> None:
        if self.cap is None:
            return

        backend = self.get_backend_name()

        fmt = self._str_to_fourcc(self.config.format)
        self.cap.set(cv2.CAP_PROP_FOURCC, fmt)
        self.cap.set(cv2.CAP_PROP_FRAME_WIDTH, self.config.width)
        self.cap.set(cv2.CAP_PROP_FRAME_HEIGHT, self.config.height)
        self.cap.set(cv2.CAP_PROP_FPS, self.config.fps)
        self.cap.set(cv2.CAP_PROP_BUFFERSIZE, self.config.buffersize)

        if self.config.auto_focus:
            self.cap.set(cv2.CAP_PROP_AUTOFOCUS, 1.0)
        else:
            self.cap.set(cv2.CAP_PROP_AUTOFOCUS, 0.0)
            self.cap.set(cv2.CAP_PROP_FOCUS, self.config.focus)

        if self.config.auto_white_balance:
            self.cap.set(cv2.CAP_PROP_AUTO_WB, 1.0)
        else:
            self.cap.set(cv2.CAP_PROP_AUTO_WB, 0.0)
            self.cap.set(
                cv2.CAP_PROP_WB_TEMPERATURE, self.config.white_balance_temperature
            )

        # Exposure control
        if backend == "V4L2":  # Linux
            self.cap.set(
                cv2.CAP_PROP_AUTO_EXPOSURE, 3.0 if self.config.auto_exposure else 1.0
            )
            if not self.config.auto_exposure:
                self.cap.set(cv2.CAP_PROP_EXPOSURE, self.config.exposure)
                self.cap.set(cv2.CAP_PROP_GAIN, self.config.gain)
        elif backend == "AVFOUNDATION":  # macOS
            # AVFoundation sometimes ignores auto_exposure, so set exposure manually
            if not self.config.auto_exposure:
                self.cap.set(cv2.CAP_PROP_EXPOSURE, self.config.exposure)
                self.cap.set(cv2.CAP_PROP_GAIN, self.config.gain)
        else:
            # Fallback for unknown backends
            self.cap.set(
                cv2.CAP_PROP_AUTO_EXPOSURE, 3.0 if self.config.auto_exposure else 1.0
            )
            if not self.config.auto_exposure:
                self.cap.set(cv2.CAP_PROP_EXPOSURE, self.config.exposure)
                self.cap.set(cv2.CAP_PROP_GAIN, self.config.gain)

    def _maybe_recover(self) -> None:
        if self._consecutive_failures >= self.MAX_CONSECUTIVE_FAILURES:
            self._consecutive_failures = 0
            self._reset_camera()

    def _reset_camera(self) -> None:
        if self.cap:
            try:
                self.cap.release()
            except Exception:
                pass
        self.cap = None

        self.connect()

    def get_backend_name(self) -> str | None:
        if not self.cap or not self.cap.isOpened():
            return None
        return self.cap.getBackendName() or "UNKNOWN"

    def update_config(self, new_config: USBCameraConfig) -> None:
        with self._lock:
            reconnect = False
            if self.config.format != new_config.format:
                reconnect = True

            self._config = new_config
            if self.cap and self.cap.isOpened():
                if reconnect:
                    self.cap.release()
                    self.connect()
                else:
                    self._apply_config()

    def connect(self) -> None:
        self.cap = cv2.VideoCapture(self.config.device_index)
        if not self.cap.isOpened():
            raise RuntimeError("Camera not available")

        self._apply_config()

    def disconnect(self) -> None:
        if self.cap and self.cap.isOpened():
            self.cap.release()
        self.cap = None

    def read(self) -> tuple[bool, Any | None, dict[str, Any] | None]:
        with self._lock:
            if self.cap is None or not self.cap.isOpened():
                self._consecutive_failures += 1
                self._maybe_recover()
                return False, None, None

            # flush stale frames to reduce latency
            for _ in range(self.config.flush_frames):
                self.cap.grab()

            # grab new frame
            if not self.cap.grab():
                self._consecutive_failures += 1
                self._maybe_recover()
                return False, None, None

            # timestamp immediately after grabbing
            t = self.ctx.clock.now_utc() if self.ctx else None

            # retrieve the frame
            ok, frame = self.cap.retrieve()
            if not ok:
                self._consecutive_failures += 1
                self._maybe_recover()
                return False, None, None

            # success, reset failure count
            self._consecutive_failures = 0
            self.set_online()

            metadata = {
                "time": t.isoformat() if t else None,
                "device": self.id,
            }

            return True, frame, metadata



from typing import Literal

from bussdcc import Device

from ...bus.gpio import GPIOInterface

from .config import DigitalOutputConfig


class DigitalOutput(Device[DigitalOutputConfig]):
    kind = "actuator"

    def __init__(self, *, id: str, config: DigitalOutputConfig):
        super().__init__(id=id, config=config)
        self.gpio: GPIOInterface | None = None

    def _level(self, state: bool) -> Literal[0, 1]:
        """Logical -> electrical"""
        if self.config.logic == "active_high":
            return 1 if state else 0
        else:
            return 0 if state else 1

    def _logical(self, level: int) -> bool:
        """Electrical -> logical"""
        if self.config.logic == "active_high":
            return level == 1
        else:
            return level == 0

    @property
    def state(self) -> bool:
        if not self.gpio:
            return False

        level = self.gpio.input(self.config.pin)
        return self._logical(level)

    def connect(self) -> None:
        if not self.ctx:
            raise RuntimeError("Device not attached")

        bus = self.ctx.runtime.devices.get(self.config.bus_id)

        if not isinstance(bus, GPIOInterface):
            raise RuntimeError(
                "DigitalOutput requires a bus that supports GPIOInterface"
            )

        self.gpio = bus.protocol()

        self.gpio.setup(
            self.config.pin,
            "out",
            initial=self._level(False),
        )

    def disconnect(self) -> None:
        if not self.gpio:
            return

        if self.config.safe_state == "on":
            self.on()
        else:
            self.off()

    def on(self) -> None:
        if self.gpio:
            self.gpio.output(self.config.pin, self._level(True))

    def off(self) -> None:
        if self.gpio:
            self.gpio.output(self.config.pin, self._level(False))

    def toggle(self) -> None:
        if self.state:
            self.off()
        else:
            self.on()



from importlib import metadata
from dataclasses import dataclass
from typing import Any


@dataclass(slots=True)
class RegistryEntry:
    name: str
    definition: Any | None
    available: bool
    error: Exception | None = None


class DeviceRegistry:
    def __init__(self) -> None:
        self.buses = self._load_entry_points("bussdcc.bus")
        self.devices = self._load_entry_points("bussdcc.device")

    def _load_entry_points(self, group: str) -> dict[str, RegistryEntry]:
        result = {}

        for ep in metadata.entry_points(group=group):
            try:
                definition = ep.load()

                result[ep.name] = RegistryEntry(
                    name=ep.name,
                    definition=definition,
                    available=True,
                    error=None,
                )

            except Exception as e:
                result[ep.name] = RegistryEntry(
                    name=ep.name,
                    definition=None,
                    available=False,
                    error=e,
                )

        return result



from typing import Literal, Any
import threading

from bussdcc import Device
from bussdcc.message import DeviceFailed
from nau7802 import NAU7802 as _NAU7802
from typed_registers import SMBusRegisterBus

from ...bus.i2c.driver import I2CBus
from .config import NAU7802Config

Channel = Literal[1, 2]


class NAU7802(Device[NAU7802Config]):
    kind = "adc"

    ADC_MAX = 8388607
    ADC_MIN = -8388608

    GAIN_MAP = {
        1: 0,
        2: 1,
        4: 2,
        8: 3,
        16: 4,
        32: 5,
        64: 6,
        128: 7,
    }

    CRS_MAP = {
        10: 0,
        20: 1,
        40: 2,
        80: 3,
        320: 7,
    }

    def __init__(self, *, id: str, config: NAU7802Config):
        super().__init__(id=id, config=config)

        self._lock = threading.Lock()
        self._current_channel: Channel | None = None

        # runtime tare offsets
        self._tare: dict[Channel, int] = {1: 0, 2: 0}

        self.device: _NAU7802 | None = None
        self.bus: I2CBus | None = None

    # ---------------------------------------------------------
    # lifecycle
    # ---------------------------------------------------------

    def connect(self) -> None:
        if not self.ctx:
            raise RuntimeError("Device not attached")

        runtime = self.ctx.runtime
        bus = runtime.devices.get(self.config.bus_id)

        if not isinstance(bus, I2CBus):
            raise RuntimeError("NAU7802 requires an I2CBus")

        protocol: SMBusRegisterBus | None = bus.protocol()
        if protocol is None:
            raise RuntimeError("I2C bus not ready")

        self.bus = bus
        self.device = _NAU7802(protocol, self.config.addr)

        with self._lock:
            self.device.initialize()
            self.device.set_gain(self.GAIN_MAP[self.config.gain])
            self.device.set_crs(self.CRS_MAP[self.config.sample_rate])

    def disconnect(self) -> None:
        self.device = None
        self.bus = None

    # ---------------------------------------------------------
    # hardware helpers
    # ---------------------------------------------------------

    def _wait_ready(self) -> bool:
        if not self.ctx or not self.device:
            return False

        timeout = self.ctx.clock.monotonic() + 0.2

        while not self.device.cycle_ready:
            self.ctx.clock.sleep(0.001)
            if self.ctx.clock.monotonic() > timeout:
                return False

        return True

    def _switch_channel(self, channel: Channel) -> bool:
        if not self.device:
            return False

        if self._current_channel == channel:
            return True

        self.device.set_channel(channel)

        # discard unstable readings after channel switch
        for _ in range(self.config.discard_samples):
            if not self._wait_ready():
                return False
            _ = self.device.adco.value

        self._current_channel = channel
        return True

    # ---------------------------------------------------------
    # raw reading
    # ---------------------------------------------------------

    def read_raw(self, channel: Channel) -> int | None:
        if not self.device:
            return None

        if not self._switch_channel(channel):
            self.set_offline()
            return None

        samples = max(1, self.config.samples)

        total = 0

        with self._lock:
            for _ in range(samples):
                if not self._wait_ready():
                    self.set_offline()
                    return None

                try:
                    value = self.device.adco.value
                except Exception as e:
                    self.set_offline(e)
                    return None

                total += value

        raw = int(total / samples)

        # detect ADC saturation
        if raw >= self.ADC_MAX or raw <= self.ADC_MIN:
            if self.ctx:
                self.ctx.emit(
                    DeviceFailed(
                        device=self.id,
                        kind=self.kind,
                        error="ADC saturation",
                    )
                )

        self.set_online()
        return raw

    # ---------------------------------------------------------
    # calibrated reading
    # ---------------------------------------------------------

    def read(self, channel: Channel) -> float | None:
        raw = self.read_raw(channel)
        if raw is None:
            return None

        value = float(raw)

        ch = self.config.channels.get(channel)
        if not ch:
            return value

        cal = ch.calibration

        value -= self._tare[channel]
        value -= cal.offset
        value *= cal.scale

        return float(value)

    # ---------------------------------------------------------
    # runtime operations
    # ---------------------------------------------------------

    def tare(self, channel: Channel) -> None:
        raw = self.read_raw(channel)
        if raw is not None:
            self._tare[channel] = raw

    def clear_tare(self, channel: Channel) -> None:
        self._tare[channel] = 0

    # ---------------------------------------------------------
    # diagnostics
    # ---------------------------------------------------------

    def status(self) -> dict[str, Any]:
        return {
            "device": self.id,
            "channel": self._current_channel,
            "tare": self._tare,
            "samples": self.config.samples,
        }
